#include "raster_util.h"
#include "clut.h"
#include "raster.h"
#include "sbuf.h"
#include <assert.h>

//
// raw data for mesh:
//
#include "mesh.cpp"


//
// simple macros to avoid including the whole vector lib
//

#define VEC_DOT_RVEC(v0, v10,v11,v12 )  ( (v0)[0]*(v10) + (v0)[1]*(v11) + (v0)[2]*(v12) )  
#define MAT_ROW_DOT_VEC(R,M,V)  (VEC_DOT_RVEC(V,(M)[0][R],(M)[1][R],(M)[2][R] ) )
#define BASE_DOT_VEC(DV,B,V)  DV[0] =  MAT_ROW_DOT_VEC(0,(B),V)+(B)[3][0],   \
                              DV[1] =  MAT_ROW_DOT_VEC(1,(B),V)+(B)[3][1],   \
                              DV[2] =  MAT_ROW_DOT_VEC(2,(B),V)+(B)[3][2];




using namespace clut_miser;
using namespace raster;


const int num_verts = sizeof(verts) / sizeof(verts[0]);
const int num_tris  = sizeof(tris) / sizeof(tris[0]);

//
// only works for meshes with <= 65535 polys
//
struct edge
{
 u16 vrs[2];
 u16 plys[2];
};
const int max_edges = 1024;


edge edges[max_edges];
int num_edges;


//
// inserts a single poly side into the edge list
//
static void add_poly_side(u16 v0,               // id of vert 0
                          u16 v1,               // id of vert 1
                          u16 poly_index,       // index of poly
                          u16 num_polys_in_mesh)// number of polys in the mesh
{
   // search for other arc
   for(int h=0;;++h)
   {
    if( h >= num_edges )
    {
     // if we make it here, the edge's partner is not registered
     // and we must create a new one.
     assert( num_edges < max_edges );

     edge & new_edge = edges[num_edges];
     ++num_edges;
   
     new_edge.vrs[0] = v0;
     new_edge.vrs[1] = v1;
     new_edge.plys[0] = poly_index;
     new_edge.plys[1] = num_polys_in_mesh; // dummy value
     break;
    }
    //
    // have we found a partner?
    //
    edge & cur_edge = edges[h];
   
    if( (cur_edge.vrs[0] == v1)&
        (cur_edge.vrs[1] == v0)&
        (cur_edge.plys[1]==num_polys_in_mesh)  )
     {
       cur_edge.plys[1] = poly_index;
       break;

     } // if found match
   }   // find arc
}

//
// crates edges for the mesh
//
void create_mesh_edges()
{
 for(int h=0;h< num_tris;++h)
 {
   // add each side of poly
   for(int v0 = 3-1,v1=0;v1<3;v0=v1,++v1)
   {
       // reverse order
       add_poly_side( tris[h][v0],tris[h][v1],u16(h), u16(num_tris) );
   }
 }  // for all polys
}


const float clip_min_x = 0.0f;
const float clip_min_y = 0.0f;
const float clip_max_x = 640.0f;
const float clip_max_y = 480.0f;

const float clip_min_z = 0.25f;


//
// draws the mesh given the object -> camera xform "xf"
//
void draw_mesh(const texture * tmap, float xf[4][3]  )
{
 //
 // screen space 2d coords
 //
 float scs[num_verts][3];
 u8    clip_flags[num_verts]; // clipping flags
 
 //
 // project all verts
 //
 for(int h=0;h<num_verts;++h)
 {
  // camera coords
  float cc[3];

  BASE_DOT_VEC(cc,xf,verts[h]);

  float z_ = 1.0f / cc[2];
  scs[h][0] = (cc[0] * z_ + 0.5f )*640.0f;
  scs[h][1] = (cc[1] * z_ + 0.5f )*640.0f;
  scs[h][2] =  z_;
  
  
  clip_flags[h]  = u8(  u8(scs[h][0] <= clip_min_x)*1 | 
                        u8(scs[h][0] >= clip_max_x)*2 | 
                        u8(scs[h][1] <= clip_min_y)*4 | 
                        u8(scs[h][1] >= clip_max_y)*8 | 
                        u8(cc[2]     <= clip_min_z)*16 ); 
                  

 }

 poly * polys[num_tris+1];
 polys[num_tris] = 0;       // this is a terminal for edges that have only 1 face
 
 
 //
 // submit all tris
 //
 for(h=0;h<num_tris;++h)
 {
  int iv0 = tris[h][0];
  int iv1 = tris[h][1];
  int iv2 = tris[h][2];
  
  u8 clip_or  = clip_flags[iv0]|clip_flags[iv1]|clip_flags[iv2];
  u8 clip_and = clip_flags[iv0]&clip_flags[iv1]&clip_flags[iv2];
  
  //
  // are we clipped?
  // note: we use agressive z clipping here
  //
  if( (clip_or&16)|(clip_and) )
  {
   polys[h] = 0;
  }
  else
  {
    vert_3s v0, v1, v2;
    v0.c[0] = scs[iv0][0];    v0.c[1] = scs[iv0][1];  v0.z_inv = scs[iv0][2];
    v1.c[0] = scs[iv1][0];    v1.c[1] = scs[iv1][1];  v1.z_inv = scs[iv1][2];
    v2.c[0] = scs[iv2][0];    v2.c[1] = scs[iv2][1];  v2.z_inv = scs[iv2][2];
    v0.shade=v1.shade = v2.shade = 0.75f;

    //
    // backface culling is done by raster 
    //
    polys[h] = submit_alt(v0,v1,v2, tmap, uvs[iv0], uvs[iv1], uvs[iv2]  ); 
  }
 }

 //
 // submit all visible edges for this mesh
 //
 for(h=0;h<num_edges;++h)
 {
  poly * p0 = polys[ edges[h].plys[0] ];
  poly * p1 = polys[ edges[h].plys[1] ];
  if( (int)p0|(int)p1 )
  {
   int v0 = edges[h].vrs[0];
   int v1 = edges[h].vrs[1];

   submit_edge( scs[v0][0], scs[v0][1],scs[v1][0], scs[v1][1],p0,p1);
  }

 }

}


extern "C" 
{
 void * __cdecl memset( void *dest, int c, size_t count );
};

//
// simple test
//
void draw_meshes( texture  * tmap, int x_res, int y_res, void * buffer, u32 buf_pitch )
{


  //
 // initialize raster
 //
 set_display_16( pixel_format_16(5,5,5) );

 //
 // set target buffer
 //
 set_surface(x_res, y_res, buffer, buf_pitch );


 //
 // make all edges
 //
 create_mesh_edges();



 float xf[4][3];

 memset(xf, 0, sizeof(xf) );

 //
 // identity
 //
 xf[0][0] = 1.0f;
 xf[1][1] = 1.0f;
 xf[2][2] = 1.0f;
 xf[3][2] = 16.0f;
 

 //
 // draw mesh a bunch of times
 //

 draw_mesh(tmap, xf );


 xf[3][2] = 19.0f;
 xf[3][1] = -1.f;
 
 draw_mesh(tmap, xf );


 xf[3][2] = 28.0f;
 xf[3][1] =  2.f;
 xf[3][0] =  2.f;
 draw_mesh(tmap, xf );


 xf[3][2] = 16.0f;
 xf[3][1] =  0.f;
 xf[3][0] = -2.f;
 draw_mesh(tmap, xf );
 

 //
 // commit to screen
 //

 raster::sweep_and_draw();

}

